#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <sys/time.h>

#include "convolve.h"

enum test_case {
	BASE_COMPLEX_REAL,
	SSE_COMPLEX_REAL,
	BASE_COMPLEX_COMPLEX,
	SSE_COMPLEX_COMPLEX,
};

static const char *get_test_name(int test)
{
	switch (test) {
	case BASE_COMPLEX_REAL:
		return "Base real-complex";
	case BASE_COMPLEX_COMPLEX:
		return "Base complex-complex";
	case SSE_COMPLEX_REAL:
		return "SSE real-complex";
	case SSE_COMPLEX_COMPLEX:
		return "SSE complex-complex";
	}

	return "";
}

static double test_convolve(int x_len, int h_len, int num, int test)
{
	int (*conv_func)(float *, int,
			 float *, int,
			 float *, int,
			 int, int, int, int);

	switch (test) {
	case BASE_COMPLEX_REAL:
		conv_func = base_convolve_real;
		break;
	case SSE_COMPLEX_REAL:
		conv_func = sse_convolve_real;
		break;
	case BASE_COMPLEX_COMPLEX:
		conv_func = base_convolve_complex;
		break;
	case SSE_COMPLEX_COMPLEX:
		conv_func = sse_convolve_complex;
		break;
	default:
		fprintf(stderr, "Invalid test case\n");
		return 0.0f;
	}

	float *in = malloc((x_len * num + h_len) * 2 * sizeof(float));
	float *h = malloc(h_len * num * 2 * sizeof(float));
	float *out = malloc(x_len * num * 2 * sizeof(float));
	const char *name = get_test_name(test);

	double elapsed;
	struct timeval tv0, tv1;

	for (int i = 0; i < h_len; i++)
		h[i] = (float) rand() / (float) RAND_MAX;

	for (int i = 0; i < num * x_len; i++)
		in[i] = (float) rand() / (float) RAND_MAX;

	gettimeofday(&tv0, NULL);

	for (int i = 0; i < num - 1; i++) {
		conv_func(&in[2 * (x_len * i + h_len)], x_len,
			  h, h_len,
			  &out[2 * (x_len * i)], x_len,
			  0, x_len, 1, 0);
	}

	gettimeofday(&tv1, NULL);

	elapsed = (tv1.tv_sec - tv0.tv_sec);
	elapsed += (tv1.tv_usec - tv0.tv_usec) / 1e6;

	printf("%s elapsed time... %f secs\n", name, elapsed);

	free(in);
	free(h);
	free(out);

	return elapsed;
}

int main(int argc, char **argv)
{
	int num;
	int x_len = 156 * 4;
	int h_len[9] = { 4, 8, 12, 16, 20, 24, 32, 64, 128 };
	double time0, time1;

	if (argc < 2)
		num = 10000;
	else
		num = atoll(argv[1]);

	if (num < 0) {
		fprintf(stderr, "Invalid length\n");
		return -1;
	}

	fprintf(stdout, "Convolving 156 * 4 samples for %i iterations\n", num);
	fprintf(stdout, "-----------------\n");
	for (int i = 0; i < 9; i++) {
		fprintf(stdout, "Filter length %i\n", h_len[i]);
		if (h_len[i] <= 20) {
			time0 = test_convolve(x_len, h_len[i],
					      num, BASE_COMPLEX_REAL);
			time1 = test_convolve(x_len, h_len[i],
					      num, SSE_COMPLEX_REAL);
			fprintf(stdout, "Quotient %f\n", time0 / time1);
			fprintf(stdout, "\n");
		}

		time0 = test_convolve(x_len, h_len[i],
				      num, BASE_COMPLEX_COMPLEX);
		time1 = test_convolve(x_len, h_len[i],
				      num, SSE_COMPLEX_COMPLEX);
		fprintf(stdout, "Quotient %f\n", time0 / time1);
		fprintf(stdout, "\n");
	}
}
