/*
 * Copyright (C) 2013 Thomas Tsou <tom@tsou.cc>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 * See the COPYING file in the main directory for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include "viterbi.h"

#define VITERBI_MAX_N		4
#define SHIFT_REG_K5 		0x0e
#define SHIFT_REG_K7 		0x3e

struct vstate {
	unsigned state;
	unsigned val;
	unsigned prev[2];
};

struct vtrellis {
	int num_states;
	struct vstate *states;
	int16_t *sums;
	int16_t *outputs;
	int16_t *metrics;
};

struct lte_vdecoder {
	int n;
	int k;
	int len;
	int recursive;
	int intrvl;
	struct vtrellis *trellis;
	int *punc;
	int16_t **paths;

	void (*branch_metric_func)(int8_t *, int16_t *, int16_t *);
	int (*path_metric_func)(int16_t *, int16_t *, int16_t *, int);
};

static unsigned gen_output(unsigned val, unsigned reg,
			   const struct lte_conv_code *code)
{
	unsigned mask;
	unsigned out[3];
	const unsigned *gen = code->gen;

	if ((gen[0] >> 7) || (gen[1] >> 7) || ((code->n > 2) & (gen[2] >> 7))) {
		fprintf(stderr, "Invalid generator polynomial\n");
		return 0;
	}

	reg = reg | (val << (code->k - 1));
	for (int i = 0; i < code->n; i++) {
		mask = reg & gen[i];
		out[i] = ((mask >> 6) & 0x01) ^
                         ((mask >> 5) & 0x01) ^
			 ((mask >> 4) & 0x01) ^
			 ((mask >> 3) & 0x01) ^
			 ((mask >> 2) & 0x01) ^
			 ((mask >> 1) & 0x01) ^
			 ((mask >> 0) & 0x01);
	}

	unsigned pack = 0;
	for (int i = 0; i < code->n; i++)
		pack |= out[i] << i;

	return pack;
}

static void gen_state_info(struct vstate *state, unsigned reg,
			   int16_t *output, const struct lte_conv_code *code)
{
	unsigned mask, out;

	switch (code->k) {
	case 5:
		mask = SHIFT_REG_K5;
		break;
	case 7:
		mask = SHIFT_REG_K7;
		break;
	default:
		fprintf(stderr, "Unsupported constraint length %i\n", code->k);
		return;
	}

	state->state = reg;

	state->prev[0] = (reg << 1 & mask) | (0 << 0);
	state->prev[1] = (reg << 1 & mask) | (1 << 0);
	state->val = (reg >> (code->k - 2)) & 0x01;

	out = gen_output(state->val, state->prev[0], code);

	/* Unpack output */
	output[0] = (((out >> 0) & 0x01) * 2 - 1) * -1;
	output[1] = (((out >> 1) & 0x01) * 2 - 1) * -1;
	output[2] = (((out >> 2) & 0x01) * 2 - 1) * -1;
	output[3] = (((out >> 3) & 0x01) * 2 - 1) * -1;
}

static void gen_recursive_state_info(struct vstate *state,
				     unsigned reg, int16_t *output,
				     const struct lte_conv_code *code)
{
	int val, val0, val1, sys = -1;
	unsigned mask, out, prev;

	switch (code->k) {
	case 5:
		mask = SHIFT_REG_K5;
		break;
	case 7:
		mask = SHIFT_REG_K7;
		break;
	default:
		fprintf(stderr, "Unsupported constraint length %i\n", code->k);
		return;
	}

	for (int i = 0; i < code->n; i++) {
		if (code->gen[i] == (1 << (code->k - 1))) {
			sys = i;
			break;
		}
	}

	if (sys < 0) {
		fprintf(stderr, "No systematic output found\n");
		return;
	}

	state->state = reg;

	/* Extract the systematic bit output - needs cleanup */
	prev = ((reg << 1) & mask) | (0 << 0);
	prev = prev & code->rgen;
	val0 = (reg >> (code->k - 2)) & 0x01;
	val1 = (prev >> 6 & 0x01) ^
	       (prev >> 5 & 0x01) ^
	       (prev >> 4 & 0x01) ^
	       (prev >> 3 & 0x01) ^
	       (prev >> 2 & 0x01) ^
	       (prev >> 1 & 0x01) ^
	       (prev >> 0 & 0x01);

	val = val0 == val1 ? 0 : 1;

	state->prev[0] = (reg << 1 & mask) | (0 << 0);
	state->prev[1] = (reg << 1 & mask) | (1 << 0);
	state->val = val; 

	out = gen_output(val0, state->prev[0], code);

	switch (sys) {
	case 0:
		out = (out & 0x0e) | ((val & 0x01) << 0);
		break;
	case 1:
		out = (out & 0x0d) | ((val & 0x01) << 1);
		break;
	case 2:
		out = (out & 0x0b) | ((val & 0x01) << 2);
		break;
	case 3:
		out = (out & 0x07) | ((val & 0x01) << 3);
		break;
	}

	output[0] = (((out >> 0) & 0x01) * 2 - 1) * -1;
	output[1] = (((out >> 1) & 0x01) * 2 - 1) * -1;
	output[2] = (((out >> 2) & 0x01) * 2 - 1) * -1;
	output[3] = 0; 
}

static struct vtrellis *generate_trellis(const struct lte_conv_code *code)
{
	int num_states = 1 << (code->k - 1);
	struct vtrellis *trellis;
	struct vstate *state;

	trellis = (struct vtrellis *) malloc(sizeof(struct vtrellis));
	trellis->num_states = num_states; 
	trellis->states = (struct vstate *)
			malloc(sizeof(struct vstate) * trellis->num_states);
	trellis->metrics = (int16_t *)
			memalign(16, sizeof(int16_t) * num_states);
	trellis->sums = (int16_t *)
			memalign(16, sizeof(int16_t) * num_states);
	trellis->outputs = (int16_t *)
			malloc(sizeof(int16_t *) * num_states * 4);

	for (unsigned i = 0; i < trellis->num_states; i++) {
		state = &trellis->states[i];
		if (code->rgen) {
			gen_recursive_state_info(state, i,
						 &trellis->outputs[4 * i],
						 code);
		} else {
			gen_state_info(state, i,
				       &trellis->outputs[4 * i],
				       code);
		}
	}

	return trellis;
}

static void free_trellis(struct vtrellis *trellis)
{
	free(trellis->outputs);
	free(trellis->sums);
	free(trellis->metrics);
	free(trellis->states);
	free(trellis);
}

static void reset_decoder(struct lte_vdecoder *dec)
{
	int num_states = dec->trellis->num_states;

	memset(dec->trellis->sums, 0, sizeof(int16_t) * num_states);
	for (int i = 0; i < dec->len; i++)
		memset(dec->paths[i], 0, sizeof(int16_t) * num_states);
}

int traceback(struct lte_vdecoder *dec, uint8_t *out)
{
	int num_states = dec->trellis->num_states;
	int len = dec->len;
	int max = 0;
	int winner = -1;
	struct vstate *state;
	int prev = 0;

	for (int i = 0; i < num_states; i++) {
		if (dec->trellis->sums[i] > max) {
			max = dec->trellis->sums[i];
			winner = i;
		}
	}

	if (winner < 0) {
		fprintf(stderr, "traceback: no valid path found\n");
		return -1;
	};

	state = &dec->trellis->states[winner];
	for (int i = len - 1; i >= 0; i--) {
		prev = dec->paths[i][state->state] + 1;

		if (!dec->recursive || !prev)
			out[i] = state->val;
		else
			out[i] = ~state->val & 0x01;

		state = &dec->trellis->states[state->prev[prev]];
	}

	return 0;
}

struct lte_vdecoder *lte_alloc_vdec(const struct lte_conv_code *code)
{
	int num_states = 1 << (code->k - 1);
	struct lte_vdecoder *dec;

	if ((code->n > 3) || (code->n < 2)) {
		fprintf(stderr, "conv: Unsupported rate %i\n", code->n);
		return NULL;
	}

	dec = (struct lte_vdecoder *) malloc(sizeof(struct lte_vdecoder));
	dec->n = code->n;
	dec->k = code->k;
	dec->len = code->len;
	dec->recursive = code->rgen ? 1 : 0;
	dec->intrvl = 32000 / (code->n * 127);
	dec->trellis = generate_trellis(code);

	dec->paths = (int16_t **) malloc(sizeof(int16_t *) * code->len);
	for (int i = 0; i < code->len; i++) {
		dec->paths[i] = (int16_t *)
				memalign(16, sizeof(int16_t) * num_states);
	}

	switch (dec->k) {
	case 5:
		dec->path_metric_func = gen_path_metrics_k5;
		switch (dec->n) {
		case 2:
			dec->branch_metric_func = gen_branch_metrics_k5_n2;
			break;
		case 3:
			dec->branch_metric_func = gen_branch_metrics_k5_n3;
			break;
		case 4:
			dec->branch_metric_func = gen_branch_metrics_k5_n4;
			break;
		default:
			fprintf(stderr, "conv: Unsupported code\n");
		}
		break;
	case 7:
		dec->path_metric_func = gen_path_metrics_k7;
		switch (dec->n) {
		case 2:
			dec->branch_metric_func = gen_branch_metrics_k7_n2;
			break;
		case 3:
			dec->branch_metric_func = gen_branch_metrics_k7_n3;
			break;
		case 4:
			dec->branch_metric_func = gen_branch_metrics_k7_n4;
			break;
		default:
			fprintf(stderr, "conv: Unsupported code\n");
		}
		break;
	}

	return dec;
}

void lte_free_vdec(struct lte_vdecoder *dec)
{
	for (int i = 0; i < dec->len; i++)
		free(dec->paths[i]);

	free(dec->paths);
	free_trellis(dec->trellis);
	free(dec);
}

/* Depuncture sequence with nagative value terminated puncturing matrix */
static int depuncture(int8_t *in, int *punc, int8_t *out, int len)
{
	int i = 0, n = 0, m = 0;

	for (i = 0; i < len; i++) {
		if ((n >= 0) && (i == punc[n])) {
			out[i] = 0;
			n++;
			if (punc[n] < 0)
				n = -1;
		} else {
			out[i] = in[m++];
		}
	}

	return 0;
}

int lte_conv_decode(struct lte_vdecoder *dec,
		    int8_t *seq, int *punc, uint8_t *out, int len)
{
	int norm = 0;
	int8_t *depunc_seq = NULL;

	if (len > dec->len) {
		fprintf(stderr, "conv: Invalid sequence length %i\n", len);
		return -1;
	}
	reset_decoder(dec);

	if (punc) {
		depunc_seq = (int8_t *)
			     malloc(len * dec->k * sizeof(int8_t));
		depuncture(seq, punc, depunc_seq, len * dec->k);
		seq = depunc_seq;
	}

	/*
	 * Propagate through the trellis with interval normalization for
 	 * overflow protection
 	 */
	for (int i = 0; i < len; i++) {
		if (i && (i % dec->intrvl))
			norm = 0;
		else
			norm = 1;

		dec->branch_metric_func(&seq[dec->n * i],
				        dec->trellis->outputs,
					dec->trellis->metrics),
		dec->path_metric_func(dec->trellis->sums,
				      dec->trellis->metrics,
				      dec->paths[i], norm);
	}

	if (depunc_seq)
		free(depunc_seq);

	return traceback(dec, out);
}
