/*
 * SSE Viterbi decoder
 *
 * Copyright (C) 2013 Thomas Tsou <tom@tsou.cc>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 * See the COPYING file in the main directory for details.
 */

#include <stdint.h>

#include <emmintrin.h>
#include <tmmintrin.h>
#include <smmintrin.h>

/* Quad-Viterbi butterfly:
 *     Compute 4-wide butterflies with 5 SSE registers. Inputs are all
 *     packed 16-bit integers in three 128-bit registers. Two intermediate
 *     registers are used and results are set in the upper 4 registers.
 *
 *     Input:
 *     M0 - Path metrics 0 (packed 16-bit integers)
 *     M1 - Path metrics 1 (packed 16-bit integers)
 *     M2 - Branch metrics (packed 16-bit integers)
 *
 *     Output:
 *     M2 - Selected and accumulated path metrics 0
 *     M4 - Selected and accumulated path metrics 1
 *     M3 - Path selections 0
 *     M1 - Path selections 1
 */
#define SSE_BUTTERFLY(M0,M1,M2,M3,M4) \
{ \
	M3 = _mm_adds_epi16(M0, M2); \
	M4 = _mm_subs_epi16(M1, M2); \
	M0 = _mm_subs_epi16(M0, M2); \
	M1 = _mm_adds_epi16(M1, M2); \
	M2 = _mm_max_epi16(M3, M4); \
	M3 = _mm_cmpgt_epi16(M3, M4); \
	M4 = _mm_max_epi16(M0, M1); \
	M1 = _mm_cmpgt_epi16(M0, M1); \
}

/* Normalize state metrics K = 5:
 *     Compute 16-wide normalization by subtracting the smallest value from
 *     all values. Inputs are 16 packed 16-bit integers across 2 XMM registers.
 *     Two intermediate registers are used and normalized results are placed
 *     in the originating locations.
 *
 *     Input:
 *     M0:1 - Path metrics 0:1 (packed 16-bit integers)
 *
 *     Output:
 *     M0:1 - Normalized path metrics 0:1
 */
#define SSE_NORMALIZE_K5(M0,M1,M2,M3) \
{ \
	M2 = _mm_minpos_epu16(M0); \
	M3 = _mm_minpos_epu16(M1); \
	M2 = _mm_min_epu16(M2, M3); \
	M2 = _mm_unpacklo_epi16(M2, M2); \
	M2 = _mm_unpacklo_epi32(M2, M2); \
	M2 = _mm_unpacklo_epi64(M2, M2); \
	M0 = _mm_subs_epi16(M0, M2); \
	M1 = _mm_subs_epi16(M1, M2); \
}

/* Normalize state metrics K = 7:
 *     Compute 64-wide normalization by subtracting the smallest value from
 *     all values. Inputs are 8 registers of state data and 4 temporary
 *     registers. Normalized results are placed in the originating locations.
 *
 *     Input:
 *     M0:7 - Path metricss 0:7 (packed 16-bit integers)
 *
 *     Output:
 *     M0:7 - Normalized path metrics 0:7
 */
#define SSE_NORMALIZE_K7(M0,M1,M2,M3,M4,M5,M6,M7,M8,M9,M10,M11) \
{ \
	M8  = _mm_min_epu16(M0, M1); \
	M9  = _mm_min_epu16(M2, M3); \
	M10 = _mm_min_epu16(M4, M5); \
	M11 = _mm_min_epu16(M6, M7); \
	M8  = _mm_min_epu16(M8, M9); \
	M9  = _mm_min_epu16(M10, M11); \
	M8  = _mm_min_epu16(M8, M9); \
	M8  = _mm_unpacklo_epi16(M8, M8); \
	M8  = _mm_unpacklo_epi32(M8, M8); \
	M8  = _mm_unpacklo_epi64(M8, M8); \
	M0  = _mm_subs_epi16(M0, M8); \
	M1  = _mm_subs_epi16(M1, M8); \
	M2  = _mm_subs_epi16(M2, M8); \
	M3  = _mm_subs_epi16(M3, M8); \
	M4  = _mm_subs_epi16(M4, M8); \
	M5  = _mm_subs_epi16(M5, M8); \
	M6  = _mm_subs_epi16(M6, M8); \
	M7  = _mm_subs_epi16(M7, M8); \
}

/* Two lane deinterleaving K = 5:
 *     Take 16 interleaved 16-bit integers and partial deinterleave to 2 packed
 *     128-bit registers, which are themselves interleaved (thus being partial
 *     deinterleaving). The operation summarized in the following graphic. 4
 *     registers are used with the lower 2 as input and upper 2 as output.
 *
 *     In   - 10101010 10101010 10101010 10101010
 *     Out  - 00000000 11111111 00000000 11111111
 *
 *     Input:
 *     M0:7 - Packed 16-bit integers
 *
 *     Output:
 *     M8:15 - Deinterleaved packed 16-bit integers
 */
#define SSE_DEINTERLEAVE_K5(M0,M1,M2,M3) \
{ \
	M0 = _mm_shufflelo_epi16(M0, _MM_SHUFFLE(3, 1, 2, 0)); \
	M0 = _mm_shufflehi_epi16(M0, _MM_SHUFFLE(3, 1, 2, 0)); \
	M1 = _mm_shufflelo_epi16(M1, _MM_SHUFFLE(3, 1, 2, 0)); \
	M1 = _mm_shufflehi_epi16(M1, _MM_SHUFFLE(3, 1, 2, 0)); \
	M0 = _mm_shuffle_epi32(M0, _MM_SHUFFLE(3, 1, 2, 0)); \
	M1 = _mm_shuffle_epi32(M1, _MM_SHUFFLE(3, 1, 2, 0)); \
	M2 = _mm_unpacklo_epi64(M0, M1); \
	M3 = _mm_unpackhi_epi64(M0, M1); \
}

/* Two lane deinterleaving K = 7:
 *     Take 64 interleaved 16-bit integers and partial deinterleave to 8 packed
 *     128-bit registers, which are themselves interleaved (thus being partial
 *     deinterleaving). The operation summarized in the following graphic. 16
 *     registers are used with the lower 8 as input and upper 8 as output.
 *
 *     In   - 10101010 10101010 10101010 10101010 ...
 *     Out  - 00000000 11111111 00000000 11111111 ...
 *
 *     Input:
 *     M0:7 - Packed 16-bit integers
 *
 *     Output:
 *     M8:15 - Deinterleaved packed 16-bit integers
 */
#define SSE_DEINTERLEAVE_K7(M0,M1,M2,M3,M4,M5,M6,M7, \
			    M8,M9,M10,M11,M12,M13,M14,M15) \
{ \
	M0  = _mm_shufflelo_epi16(M0, _MM_SHUFFLE(3, 1, 2, 0)); \
	M0  = _mm_shufflehi_epi16(M0, _MM_SHUFFLE(3, 1, 2, 0)); \
	M1  = _mm_shufflelo_epi16(M1, _MM_SHUFFLE(3, 1, 2, 0)); \
	M1  = _mm_shufflehi_epi16(M1, _MM_SHUFFLE(3, 1, 2, 0)); \
	M2  = _mm_shufflelo_epi16(M2, _MM_SHUFFLE(3, 1, 2, 0)); \
	M2  = _mm_shufflehi_epi16(M2, _MM_SHUFFLE(3, 1, 2, 0)); \
	M3  = _mm_shufflelo_epi16(M3, _MM_SHUFFLE(3, 1, 2, 0)); \
	M3  = _mm_shufflehi_epi16(M3, _MM_SHUFFLE(3, 1, 2, 0)); \
	M4  = _mm_shufflelo_epi16(M4, _MM_SHUFFLE(3, 1, 2, 0)); \
	M4  = _mm_shufflehi_epi16(M4, _MM_SHUFFLE(3, 1, 2, 0)); \
	M5  = _mm_shufflelo_epi16(M5, _MM_SHUFFLE(3, 1, 2, 0)); \
	M5  = _mm_shufflehi_epi16(M5, _MM_SHUFFLE(3, 1, 2, 0)); \
	M6  = _mm_shufflelo_epi16(M6, _MM_SHUFFLE(3, 1, 2, 0)); \
	M6  = _mm_shufflehi_epi16(M6, _MM_SHUFFLE(3, 1, 2, 0)); \
	M7  = _mm_shufflelo_epi16(M7, _MM_SHUFFLE(3, 1, 2, 0)); \
	M7  = _mm_shufflehi_epi16(M7, _MM_SHUFFLE(3, 1, 2, 0)); \
	M0  = _mm_shuffle_epi32(M0, _MM_SHUFFLE(3, 1, 2, 0)); \
	M1  = _mm_shuffle_epi32(M1, _MM_SHUFFLE(3, 1, 2, 0)); \
	M2  = _mm_shuffle_epi32(M2, _MM_SHUFFLE(3, 1, 2, 0)); \
	M3  = _mm_shuffle_epi32(M3, _MM_SHUFFLE(3, 1, 2, 0)); \
	M4  = _mm_shuffle_epi32(M4, _MM_SHUFFLE(3, 1, 2, 0)); \
	M5  = _mm_shuffle_epi32(M5, _MM_SHUFFLE(3, 1, 2, 0)); \
	M6  = _mm_shuffle_epi32(M6, _MM_SHUFFLE(3, 1, 2, 0)); \
	M7  = _mm_shuffle_epi32(M7, _MM_SHUFFLE(3, 1, 2, 0)); \
	M8  = _mm_unpacklo_epi64(M0, M1); \
	M9  = _mm_unpackhi_epi64(M0, M1); \
	M10 = _mm_unpacklo_epi64(M2, M3); \
	M11 = _mm_unpackhi_epi64(M2, M3); \
	M12 = _mm_unpacklo_epi64(M4, M5); \
	M13 = _mm_unpackhi_epi64(M4, M5); \
	M14 = _mm_unpacklo_epi64(M6, M7); \
	M15 = _mm_unpackhi_epi64(M6, M7); \
}

/* Branch metric unit for 16-state trellis (k = 5) up to 1/4 rate */
static void _sse_branch_metrics_k5(int8_t *restrict val,
				   int16_t *restrict outputs,
				   int16_t *restrict metrics)
{
	__m128i m0, m1, m2, m3, m4;

	/* Load and expand 8-bit input out to 16-bits */
	m4 = _mm_loadl_epi64((__m128i *) val);
	m4 = _mm_cvtepi8_epi16(m4);

	/* Load path symbols */
	m0 = _mm_load_si128((__m128i *) &outputs[0]);
	m1 = _mm_load_si128((__m128i *) &outputs[8]);
	m2 = _mm_load_si128((__m128i *) &outputs[16]);
	m3 = _mm_load_si128((__m128i *) &outputs[24]);

	/* Multiply */
	m0 = _mm_mullo_epi16(m0, m4);
	m1 = _mm_mullo_epi16(m1, m4);
	m2 = _mm_mullo_epi16(m2, m4);
	m3 = _mm_mullo_epi16(m3, m4);

	/* Accumulate */
	m0 = _mm_hadd_epi16(m0, m1);
	m1 = _mm_hadd_epi16(m2, m3);
	m0 = _mm_hadd_epi16(m0, m1);

	_mm_store_si128((__m128i *) metrics, m0);
}

/* Branch metric unit for 64-state trellis (k = 7) up to 1/4 rate */
static void _sse_branch_metrics_k7(int8_t *restrict val,
				   int16_t *restrict outputs,
				   int16_t *restrict metrics)
{
	__m128i m0, m1, m2, m3, m4, m5, m6, m7, m8;

	/* Load and expand 8-bit input out to 16-bits */
	m8 = _mm_loadl_epi64((__m128i *) val);
	m8 = _mm_cvtepi8_epi16(m8);

	for (int i = 0; i < 2; i++) {
		/* Load path symbols */
		m0 = _mm_load_si128((__m128i *) &outputs[64 * i + 0]);
		m1 = _mm_load_si128((__m128i *) &outputs[64 * i + 8]);
		m2 = _mm_load_si128((__m128i *) &outputs[64 * i + 16]);
		m3 = _mm_load_si128((__m128i *) &outputs[64 * i + 24]);
		m4 = _mm_load_si128((__m128i *) &outputs[64 * i + 32]);
		m5 = _mm_load_si128((__m128i *) &outputs[64 * i + 40]);
		m6 = _mm_load_si128((__m128i *) &outputs[64 * i + 48]);
		m7 = _mm_load_si128((__m128i *) &outputs[64 * i + 56]);
        
		/* Multiply */
		m0 = _mm_mullo_epi16(m0, m8);
		m1 = _mm_mullo_epi16(m1, m8);
		m2 = _mm_mullo_epi16(m2, m8);
		m3 = _mm_mullo_epi16(m3, m8);
		m4 = _mm_mullo_epi16(m4, m8);
		m5 = _mm_mullo_epi16(m5, m8);
		m6 = _mm_mullo_epi16(m6, m8);
		m7 = _mm_mullo_epi16(m7, m8);
        
		/* Accumulate */
		m0 = _mm_hadd_epi16(m0, m1);
		m1 = _mm_hadd_epi16(m2, m3);
		m2 = _mm_hadd_epi16(m4, m5);
		m3 = _mm_hadd_epi16(m6, m7);
		m0 = _mm_hadd_epi16(m0, m1);
		m2 = _mm_hadd_epi16(m2, m3);
        
		_mm_store_si128((__m128i *) &metrics[16 * i + 0], m0);
		_mm_store_si128((__m128i *) &metrics[16 * i + 8], m2);
	}
}

/* Path metric unit for 16-state trellis (k = 5)
 * This works by deinterleaving even / odd state values into separate 128-bit
 * SSE registers in order to setup a single 8-wide parallel butterfly pass.
 */
static int _sse_path_metrics_k5(int16_t *restrict sums,
				int16_t *restrict metrics,
				int16_t *restrict paths,
				int norm)
{
	__m128i m0, m1, m2, m3, m4, m5, m6;

	/* Load branch metrics and accumulated path matrics */
	m0 = _mm_load_si128((__m128i *) &sums[0]);
	m1 = _mm_load_si128((__m128i *) &sums[8]);
	m2 = _mm_load_si128((__m128i *) metrics);

	/* Deinterleave into even and odd packed registers */
	SSE_DEINTERLEAVE_K5(m0, m1, m3, m4)

	/* Butterflies: 0-7 */ 
	SSE_BUTTERFLY(m3, m4, m2, m5, m6)

	if (norm)
		SSE_NORMALIZE_K5(m2, m6, m0, m1)

	_mm_store_si128((__m128i *) &sums[0], m2);
	_mm_store_si128((__m128i *) &sums[8], m6);
	_mm_store_si128((__m128i *) &paths[0], m5);
	_mm_store_si128((__m128i *) &paths[8], m4);

	return 0;
}

/* Path metrics for 64-state trellis (k = 7)
 *     32 butterflies are computed in 2 sets of 16 - corresponding to two SSE
 *     register widths of 128-bits (8 x 16-bit integers). Run a large initial
 *     3-stage shuffle-unpack sequence to arrange even-odd state variables
 *     in an 8-pack configuration.
 */
static int _sse_path_metrics_k7(int16_t *restrict sums,
				int16_t *restrict metrics,
				int16_t *restrict paths,
				int norm)
{
	__m128i m0, m1, m2, m3, m4, m5, m6, m7;
	__m128i m8, m9, m10, m11, m12, m13, m14, m15;

	/* Load accumulated path matrics */
	m0 = _mm_load_si128((__m128i *) &sums[0]);
	m1 = _mm_load_si128((__m128i *) &sums[8]);
	m2 = _mm_load_si128((__m128i *) &sums[16]);
	m3 = _mm_load_si128((__m128i *) &sums[24]);
	m4 = _mm_load_si128((__m128i *) &sums[32]);
	m5 = _mm_load_si128((__m128i *) &sums[40]);
	m6 = _mm_load_si128((__m128i *) &sums[48]);
	m7 = _mm_load_si128((__m128i *) &sums[56]);

	/* Deinterleave into even and odd packed registers */
	SSE_DEINTERLEAVE_K7(m0, m1, m2, m3 ,m4 ,m5, m6, m7,
			    m8, m9, m10, m11, m12, m13, m14, m15)

	/* Load branch metrics */
	m0 = _mm_load_si128((__m128i *) &metrics[0]);
	m1 = _mm_load_si128((__m128i *) &metrics[8]);
	m2 = _mm_load_si128((__m128i *) &metrics[16]);
	m3 = _mm_load_si128((__m128i *) &metrics[24]);

	/* Butterflies: 0-15 */ 
	SSE_BUTTERFLY(m8, m9, m0, m4, m5)
	SSE_BUTTERFLY(m10, m11, m1, m6, m7)

	_mm_store_si128((__m128i *) &paths[0], m4);
	_mm_store_si128((__m128i *) &paths[8], m6);
	_mm_store_si128((__m128i *) &paths[32], m9);
	_mm_store_si128((__m128i *) &paths[40], m11);

	/* Butterflies: 17-31 */
	SSE_BUTTERFLY(m12, m13, m2, m8, m10)
	SSE_BUTTERFLY(m14, m15, m3, m4, m6)

	_mm_store_si128((__m128i *) &paths[16], m8);
	_mm_store_si128((__m128i *) &paths[24], m4);
	_mm_store_si128((__m128i *) &paths[48], m13);
	_mm_store_si128((__m128i *) &paths[56], m15);

	/* Delay state variable storage in case we need to normalize */
	if (norm) {
		SSE_NORMALIZE_K7(m0, m5, m1, m7, m2, m10,
				 m3, m6, m4, m8, m9, m11)
	}

	_mm_store_si128((__m128i *) &sums[0], m0);
	_mm_store_si128((__m128i *) &sums[8], m1);
	_mm_store_si128((__m128i *) &sums[16], m2);
	_mm_store_si128((__m128i *) &sums[24], m3);
	_mm_store_si128((__m128i *) &sums[32], m5);
	_mm_store_si128((__m128i *) &sums[40], m7);
	_mm_store_si128((__m128i *) &sums[48], m10);
	_mm_store_si128((__m128i *) &sums[56], m6);

	return 0;
}

/* Generate branch metrics:
 *     Zero pad input values up to 4 and replicate so to pack two sets of the
 *     same nput value in a single 128-bit register. We use the same branch
 *     metric unit for all values of N up to 4. The metric unit will take N
 *     input values and unpack, expand, and repack into a SSE register
 *     at 2 x 4 x 16-bits.
 */
void gen_branch_metrics_k5_n2(int8_t *val, int16_t *outputs, int16_t *metrics)
{
	int8_t _val[16] = {
		val[0], val[1], 0, 0,
		val[0], val[1], 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
	};

	_sse_branch_metrics_k5(_val, outputs, metrics);
}

void gen_branch_metrics_k5_n3(int8_t *val, int16_t *outputs, int16_t *metrics)
{
	int8_t _val[16] = {
		val[0], val[1], val[2], 0,
		val[0], val[1], val[2], 0,
		0, 0, 0, 0, 0, 0, 0, 0,
	};

	_sse_branch_metrics_k5(_val, outputs, metrics);
}

void gen_branch_metrics_k5_n4(int8_t *val, int16_t *outputs, int16_t *metrics)
{
	int8_t _val[16] = {
		val[0], val[1], val[2], val[3],
		val[0], val[1], val[2], val[3],
		0, 0, 0, 0, 0, 0, 0, 0,
	};

	_sse_branch_metrics_k5(_val, outputs, metrics);
}

void gen_branch_metrics_k7_n2(int8_t *val, int16_t *outputs, int16_t *metrics)
{
	int8_t _val[16] = {
		val[0], val[1], 0, 0,
		val[0], val[1], 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
	};

	_sse_branch_metrics_k7(_val, outputs, metrics);
}

void gen_branch_metrics_k7_n3(int8_t *val, int16_t *outputs, int16_t *metrics)
{
	int8_t _val[16] = {
		val[0], val[1], val[2], 0,
		val[0], val[1], val[2], 0,
		0, 0, 0, 0, 0, 0, 0, 0,
	};

	_sse_branch_metrics_k7(_val, outputs, metrics);
}

void gen_branch_metrics_k7_n4(int8_t *val, int16_t *outputs, int16_t *metrics)
{
	int8_t _val[16] = {
		val[0], val[1], val[2], val[3],
		val[0], val[1], val[2], val[3],
		0, 0, 0, 0, 0, 0, 0, 0,
	};

	_sse_branch_metrics_k7(_val, outputs, metrics);
}

int gen_path_metrics_k5(int16_t *sums, int16_t *metrics, int16_t *paths, int norm)
{
	return _sse_path_metrics_k5(sums, metrics, paths, norm);
}

int gen_path_metrics_k7(int16_t *sums, int16_t *metrics, int16_t *paths, int norm)
{
	return _sse_path_metrics_k7(sums, metrics, paths, norm);
}
